from piku.core import config, modules


def install_command(args):
    deps = config.get('tool.piku.dependencies')
    for name, version in deps.items():

        # determind module source
        if version.startswith('file:'):
            source = version
        else:
            source = f'{name}@{version}'

        # decode module
        module, type, path, version = modules.decode(source)

        # copy or download module
        print(f'Acquiring {module}...')
        if not modules.aquire(path):
            print(f'Unable to aquire requested module: {module}')
            return



from piku.commands.version import get_version
from piku.core import modules, config, utils


def info_command(args):
    print(f'Piku Version: {get_version()}')
    print(f'Bundle directory: {config.bundle_path}')
    print(f'Backup directory: {config.backup_path}')
    print(f'Deploy directory: {config.deploy_path}')

    if args.clear_cache:
        print('Clearing module cache...')
        try:
            utils.remove(config.bundle_path)
        except FileNotFoundError:
            pass
        print('Done')

    if args.modules:
        index = modules.index(config.get('tool.piku.circuitpython'))
        print('Availiable modules...')
        for module in index.keys():
            print(f' * {module}')
        print('Done')



import os
import toml
from appdirs import user_data_dir
from piku.core.utils import nset, nget, ndel


defaults = {
    'tool':
    {
        'piku': {
        'source': './project'
        }
    }
}

config_path = os.path.join(os.getcwd(), 'pyproject.toml')
data_path = user_data_dir('piku', 'piku')
bundle_path = os.path.join(data_path, 'bundle')
backup_path = os.path.join(data_path, 'backup')
deploy_path = os.path.join(data_path, 'deploy')


def load():
    return toml.load(config_path)

def save(config):
    with open(config_path, 'w') as file:
        toml.dump(config, file)

def get(dotpath):
    config = load()
    assert dotpath.startswith('tool.piku')
    keys = dotpath.split('.')
    return nget(config, keys, nget(defaults, keys))

def set(dotpath, value):
    config = load()
    assert dotpath.startswith('tool.piku')
    keys = dotpath.split('.')
    nset(config, keys, value)
    save(config)

def remove(dotpath):
    config = load()
    assert dotpath.startswith('tool.piku')
    keys = dotpath.split('.')
    ndel(config, keys)
    save(config)



import os
from jinja2 import Template
from piku.core import utils
from piku.commands.version import get_version


def template(path, context):
    with open(path, 'r') as f:
        template = Template(f.read())
    with open(path, 'w') as f:
        f.write(template.render(**context))

def create_command(args):
    # build template context
    project_path = f'./{args.project}'
    toml_path = os.path.join(project_path, 'pyproject.toml')
    readme_path = os.path.join(project_path, 'README.md')
    context = {
        'project': args.project,
        'piku': get_version(),
        'bundle': '7',
        'device': '',
        'serial': ''
    }

    # show piku version
    print(f'Piku v{context["piku"]}')

    # check that path doesnt exist
    if os.path.exists(project_path):
        print(f'Unable to create project: directory {project_path} already exists.')
        return

    # create project from template
    src = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../template')
    utils.copy(src, project_path, contents=True)

    # template files
    template(toml_path, context)
    template(readme_path, context)

    print('Done.')

    # in the future allow loading an example and dependencies for a specific board

    # # select board
    # print('Please enter the name of your CircuitPython board:')
    # context['board'] = input().strip()

    # # select device
    # print('Please enter the usb drive path of your device CircuitPython enabled device:')
    # context['device'] = input().strip()
    #
    # # select serial port
    # print('Please enter the serial port for your CircuitPython device:')
    # context['serial'] = input().strip()



#!/usr/bin/env python3
import sys
import argparse
import traceback
from piku import commands


def main():
    # create parser
    parser = argparse.ArgumentParser(description='')
    subparsers = parser.add_subparsers(dest='parser')
    parsers = {}

    # create command
    p = subparsers.add_parser('create', help='create new CircuitPython project')
    p.set_defaults(cmd=commands.create_command)
    p.add_argument('project', help='project name')

    # add command
    p = subparsers.add_parser('add', help='download and add module to project')
    p.set_defaults(cmd=commands.add_command)
    p.add_argument('module', help='module name [neopixel, file:/path/to/module.mpy]')

    # remove command
    p = subparsers.add_parser('remove', help='remove module from project')
    p.set_defaults(cmd=commands.remove_command)
    p.add_argument('module', help='module name [neopixel, my_module_name]')

    # install command
    p = subparsers.add_parser('install', help='install project dependencies')
    p.set_defaults(cmd=commands.install_command)

    # serial command
    p = subparsers.add_parser('serial', help='connect usb serial port of device')
    p.set_defaults(cmd=commands.serial_command)
    p.add_argument('-s', '--serial', default=None, help='serial port to connect to')

    # deploy command
    p = subparsers.add_parser('deploy', help='deploy project to device')
    p.set_defaults(cmd=commands.deploy_command)
    p.add_argument('-d', '--device', default=None, help='path of device to deploy to project to')
    p.add_argument('-y', '--yes', action='store_true', help='deploy to discovered device without confirmation dialog')

    # version command
    p = subparsers.add_parser('version', help='show piku version')
    p.set_defaults(cmd=commands.version_command)

    # info command
    p = subparsers.add_parser('info', help='show additional piku information')
    p.set_defaults(cmd=commands.info_command)
    p.add_argument('--clear-cache', action='store_true', help='clear module cache')
    p.add_argument('-m', '--modules', action='store_true', help='list availiable modules')

    # parse and execute
    args = parser.parse_args()
    if hasattr(args, 'cmd'):
        try:
            args.cmd(args)
        except Exception as exc:
            print(traceback.format_exc(), end='')
            print('Failed.')
        sys.exit()
    else:
        parsers.get(args.parser, parser).print_help()


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        print('')



from serial import Serial
from serial.serialutil import SerialException
from serial.tools import list_ports
from serial.tools.miniterm import Miniterm
from piku.core import config


def default():
    ports = list_ports.comports()
    if not ports: return None
    return ports[0].device

def serial_command(args):
    baud = 115200
    using_default = False
    port = args.serial
    if not port:
        using_default = True
        port = default()

    # print help text
    print(r'================================================================================')
    print(f'Connecting to {"DEFAULT" if using_default else "SPECIFIED"} serial device {port} using baud {baud}...')
    print(r'  * ctrl-x: close terminal')
    print(r'  * ctrl-c: enter python repl on device')
    print(r'  * ctrl-d: exit python repl on device and reload')
    print(r'================================================================================')

    # confirm that a port was configured or a default port was found
    if not port:
        print('Failed to find the specified serial port or find a default serial port to use.')
        return

    # create serial port connection
    try:
        serial = Serial(port, baud)
    except SerialException:
        if using_default:
            print(f'Failed to open the default serial port {port}')
        else:
            print(f'Failed to open the configured serial port {port}')
        return

    # start terminal
    miniterm = Miniterm(serial, echo=False, eol='crlf')
    miniterm.exit_character = chr(0x18)  # GS/CTRL+])
    miniterm.menu_character = chr(0x14)  # Menu: CTRL+T
    miniterm.raw = False
    miniterm.set_rx_encoding('UTF-8')
    miniterm.set_tx_encoding('UTF-8')
    miniterm.start()

    # wait for terminal exit
    try:
        miniterm.join(True)
    except KeyboardInterrupt:
        pass
    miniterm.join()
    miniterm.close()



import shutil
import platform
from subprocess import check_output
from piku.core import config
from piku.core.sync import sync
from piku.core.backup import backup



def has_correct_size(path):
    total, used, free = shutil.disk_usage(path)
    return 0 < total < 3E6

def has_correct_label(path):
    if platform.system() == 'Windows':
        drive = path.split(':')[0]
        label = check_output(f'cmd /c vol {drive}:'.split()).decode().split('\r\n')[0].split(' ').pop()
        return 'circuitpy' in label.lower()
    return 'circuitpy' in path.lower()

def find_device_path():
    if platform.system() == 'Windows':
        output = check_output('wmic logicaldisk where drivetype=2 get DeviceId , VolumeName'.split()).decode()
        drives = [line.split(' ')[0] for line in output.split('\r\n') if 'circuitpy' in line.lower()]
        return drives[0] if drives else None
    if platform.system() == 'Linux':
        output = check_output('lsblk -l -o mountpoint,label,rm'.split()).decode()
        lines = [line.split() for line in output.split('\n')]
        drives = [line[0] for line in lines if len(line) == 3 and 'circuitpy' in line[1].lower() and line[2] == "1"]
        return drives[0] if drives else None
    return None


def deploy_command(args):
    # get device
    source = config.get('tool.piku.source')
    device = args.device or find_device_path()

    # check that we have a device found or specified
    if not device:
        print('Unable find a device and deploy, please specify a device to deploy to.')
        return

    # check that device size and name are as expected to reduce chances of loading onto wrong device
    if not has_correct_size(device):
        print('Refusing to deploy, specified CircuitPython drive is larger than expected (~2MB).')
        return
    if not has_correct_label(device):
        print('Refusing to deploy, expected device to have "circuitpy" in path.')
        return

    # confirm deploy
    if not args.yes:
        print(f'Are you sure you want to deploy to device: {device}?')
        print('WARNING THIS WILL REMOVE ALL OTHER FILES FROM THE DEVICE! PLEASE BE CAREFUL!')
        response = input('Are you sure? [y/n] ').lower()
        if response not in ['y', 'yes']:
            print('Exiting')
            return

    # backup device files before deploy
    backup_path = config.backup_path
    print(f'Backing up device files from {device} to {backup_path}...')
    backup(device, backup_path)

    # synchronize files to device
    print(f'Deploying project to device {device}...')
    sync(source, device, exclude=['boot_out.txt', '.*'], verbosity=1)

    print('Done')



import os
import fnmatch
from piku.core.utils import checksum, tree, copy, remove

# helpers
def different(a, b):
    if os.path.isdir(a) and os.path.isdir(b):
        return False
    if os.path.isfile(a) and os.path.isfile(b):
        return os.path.getsize(a) != os.path.getsize(b) or checksum(a) != checksum(b)
    return True

def ignored(path, patterns):
    for pattern in patterns:
        if fnmatch.filter([path], pattern):
            return True
    return False


# synchronize
def sync(src_dir, dst_dir, exclude=None, verbosity=1):
    src_tree = tree(src_dir)
    dst_tree = tree(dst_dir)
    to_rm = dst_tree - src_tree
    to_ignore = {os.path.join(dst_dir, i) for i in (exclude or [])}
    changes_detected = False

    # create any missing dir and copy over missing files (sorted so dirs are created files)
    for path in sorted(src_tree):
        src = os.path.join(src_dir, path)
        dst = os.path.join(dst_dir, path)
        if ignored(dst, to_ignore):
            if verbosity > 1: print(f'* Ignoring copy {dst}')
        elif different(src, dst):
            if verbosity > 0: print(f'* Copying {src} to {dst}')
            copy(src, dst, recursive=False)
            changes_detected = True
        else:
            if verbosity > 1: print(f'* Files {src} {dst} are the same')

    # remove files that are in destination but not in source (reverse sorted so files rm before dirs)
    for path in reversed(sorted(to_rm)):
        full_path = os.path.join(dst_dir, path)
        if ignored(full_path, to_ignore):
            if verbosity > 1: print(f'* Ignoring remove {full_path}')
        else:
            if verbosity > 0: print(f'* Removing {full_path}')
            remove(full_path, recursive=False)
            changes_detected = True

    if verbosity > 0 and not changes_detected:
        print('Nothing to deploy, no changes were detected.')



import os
from piku.core import config, utils
from piku.core.index import index


# find a module
def find(module, bundle=None):
    bundle = bundle or config.get('tool.piku.circuitpython')
    return index(bundle).get(module)

# suggest a module from index lexically similar to the one provided
def suggest(module, bundle=None):
    bundle = bundle or config.get('tool.piku.circuitpython')
    return utils.similar(module, index(bundle).keys())

# copy a module from source to project library
def aquire(source, project_path=None):
    project_path = project_path or config.get('tool.piku.source')
    library_path = os.path.join(project_path, 'lib')
    if os.path.isdir(project_path):
        os.makedirs(library_path, exist_ok=True)
    return utils.copy(source, library_path)

# remove a module from project library
def remove(module, library_path=None):
    library_path = library_path or os.path.join(config.get('tool.piku.source'), 'lib')
    for path in os.listdir(library_path):
        if path in [module, f'{module}.mpy']:
            utils.remove(os.path.join(library_path, path))
            return True
    return False

# decode module name and source from request module path
def decode(module):

    # module from local file
    if module.startswith('file:'):
        path = module[5:]
        name = os.path.splitext(os.path.basename(path))[0]
        type = 'file'
        version = module

    # module from index via semver
    else:
        bundle = config.get('tool.piku.circuitpython')
        parts = module.split('@')
        name = parts[0].lower()
        type = 'index'
        path = find(name)
        if len(parts) > 1:
            version = parts[1]
        else:
            version = f'~{bundle}'
    return (name, type, path, version)



import os
import shutil
from time import time
from piku.core import config


def backup(src, dst):
    # backup drive contents
    os.makedirs(dst, exist_ok=True)
    timestamp = int(time() * 1000)
    backup_path = os.path.join(dst, str(timestamp))
    shutil.copytree(src, backup_path)

    # remove old backups (keep the most recent 10)
    dirs = os.listdir(dst)
    dirs.sort(key=lambda x: -int(x))
    for dir in dirs[10:]:
        shutil.rmtree(os.path.join(dst, dir))



import os
import json
import zipfile
from piku.core import config, utils


bundles = {
    '7': {
        'official': 'https://github.com/adafruit/Adafruit_CircuitPython_Bundle/releases/download/20220131/adafruit-circuitpython-bundle-7.x-mpy-20220131.zip',
        'community': 'https://github.com/adafruit/CircuitPython_Community_Bundle/releases/download/20220113/circuitpython-community-bundle-7.x-mpy-20220113.zip'
    }
}

# create index of known offical and community modules
def index(bundle):
    # check that we have sources for this bundle
    if bundle not in bundles:
        return None

    # construct paths
    bundle_cache_path = os.path.join(config.bundle_path, bundle)
    bundle_index_path = os.path.join(bundle_cache_path, 'index.json')
    official_zip_url = bundles[bundle]['official']
    official_zip_file = official_zip_url.split('/')[-1]
    official_zip_path = os.path.join(bundle_cache_path, official_zip_file)
    official_bundle_path = os.path.join(bundle_cache_path, official_zip_file.replace('.zip', ''))
    official_lib_path = os.path.join(official_bundle_path, 'lib')
    community_zip_url = bundles[bundle]['community']
    community_zip_file = community_zip_url.split('/')[-1]
    community_zip_path = os.path.join(bundle_cache_path, community_zip_file)
    community_bundle_path = os.path.join(bundle_cache_path, community_zip_file.replace('.zip', ''))
    community_lib_path = os.path.join(community_bundle_path, 'lib')

    # load index
    try:
        with open(bundle_index_path, 'r') as f:
            return json.load(f)
    except FileNotFoundError:
        pass

    # create bundle cache dir
    os.makedirs(bundle_cache_path, exist_ok=True)

    # download bundle zip files
    if not os.path.exists(official_zip_path):
        print ('Downloading official bundle...')
        utils.download(bundles[bundle]['official'], official_zip_path)
        print('Done')
    if not os.path.exists(community_zip_path):
        print ('Downloading community bundle...')
        utils.download(bundles[bundle]['community'], community_zip_path)
        print('Done')

    # extract bundles
    with zipfile.ZipFile(official_zip_path, 'r') as zip:
        zip.extractall(bundle_cache_path)
    with zipfile.ZipFile(community_zip_path, 'r') as zip:
        zip.extractall(bundle_cache_path)

    # build index
    idx = {}
    for module in os.listdir(official_lib_path):
        name = module.replace('.mpy', '')
        idx[name] = os.path.join(official_lib_path, module)
    for module in os.listdir(community_lib_path):
        name = module.replace('.mpy', '')
        idx[name] = os.path.join(community_lib_path, module)
    with open(bundle_index_path, 'w') as f:
        json.dump(idx, f, indent=2)

    # return index
    return idx



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['piku', 'piku.commands', 'piku.core', 'piku.template.project']

package_data = \
{'': ['*'], 'piku': ['template/*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'appdirs>=1.4.4,<2.0.0',
 'pyserial>=3.5,<4.0',
 'requests>=2.27.1,<3.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['piku = piku.main:main']}

setup_kwargs = {
    'name': 'piku',
    'version': '0.1.6',
    'description': '',
    'long_description': '# Piku\nPiku is small command line utility for managing CircuitPython projects\n\nThe purpose of this project is to make creating a CircuitPython project, installing packages, deploying, and connecting to a CircuitPython device easy to do from the command line.\n\n\n# Warning\nThis tool is in very early development and needs testing! Please be careful when deploying and make sure you are only deploying your CircuitPython device!  Use at your own risk.\n\n\n# Getting Started\n\n### Installation\nPiku has been lightly tested on Linux and Windows it may also work on macOS.  I don\'t have an Apple computer so contributions welcome! After installation you can learn about Piku or any command line arguments or flags type `piku -h` or `piku <command> -h` or the documentation here.\n\n##### Windows\nTo install Piku in Windows please install Python 3.8 or greater from the windows store or the official python website.  Then install using pip:\n\n```\npip install piku\n```\n\nAfter Piku is installed you should be able to run Piku from the command line.  You can test this by typing `piku version`.\n\n##### Linux\nTo install Piku in Linux, make sure you have Python 3.8 or greater and install using pip3.\n\n```\npip install --user piku\n```\n\nAfter Piku is installed you should be able to run Piku from the command line.  You can test this by typing `piku version`.\n\nSome linux computers do not have the default pip user bin directory included in the PATH.  You may add this directory to your PATH or install without the `--user` argument.\nhttps://unix.stackexchange.com/questions/612779/warning-not-on-path-when-i-tried-to-install-python-extensions-im-new-to-linu\n\n\nAfter installation if your user does not have permissions to use the serial port, you may need to add your user to the `dialout` group.\nhttps://askubuntu.com/questions/58119/changing-permissions-on-serial-port#answer-58122\n\n\n#### MacOS\n\nHelp wanted!  The process should be similar to Linux.\n\n\n### Preparing your Device\n\nBefore creating a project you must have CircuitPython installed on your device, and have your device serial and USB drivers installed.  Please check the CircuitPython website for instructions or the documentation of the board you have purchased.  When your done you should be able to see your drive mounted as a USB drive named `CIRCUITPY`.\n\nhttps://learn.adafruit.com/welcome-to-circuitpython/installing-circuitpython\n\n\n### Creating a Project\n\nTo create a new Piku project from the command line type:\n\n```\npiku create example\n```\n\nThis will create a new directory with the name of your project and a few folders and files inside. After you have created a project to use Piku, enter the directory of the project you just created to use Piku:\n\n```\ncd example\n```\n\n\n### Deploying your Project\nAfter you have created a project and modified the main `code.py` file you can deploy your project to a connected CircuitPython device.  To deploy your project find the path of your `CIRCUITPY` UDB drive.  Then type:\n\n```\npiku deploy -d <path of your device>\n```\n\n***WARNING!!***  \nDeploying will remove other files from your device.  Piku attempts to check that the device is actually a CircuitPython device, and backup your old files, but you still need to be very careful.\n\nYou can also let Piku find your device by running deploy with no device argument:\n\n```\npiku deploy\n```\n\nAfter you have confirmed multiple times that you are deploying to the correct device you can also live on the wild side and skip the confirmation dialog using the `-y` command line argument.  Please be careful.\n\nIf changes have been made in your project code, the CircuitPython device should automatically detect and change files and reload.\n\n\n### Connecting to your Device\n\nYou can also connect to your CircuitPython device\'s serial port using Piku.  To do this just user the serial command from your Piku project directory:\n\n```\npiku serial\n```\n\nIf you are unable to connect, please confirm that you have the serial drivers for your device installed and you have permission to use the serial port (see installation instructions).  If you know the serial port, or Piku is connecting to the wront port you can also try specifying it directly via the `-s` command line flag.\n\nOnce connected you can exit by typing `ctrl-x`, enter the CircuitPython REPL by hitting `ctrl-c` and `ctrl-d` to exit the CircuitPython REPL.\n\n\n### Managing CircuitPython Modules/Libraries\n\nYou can easily download and add CircuitPython modules from the official Bundle or Community bundle using the command.  For example to download and add the `neopixel` module you would type:\n\n```\npiku add neopixel\n```\n\nThe specified module will be downloaded and added to your `lib` folder and your `pyproject.toml` file. You can also remove this module by typing:\n\n```\npiku remove neopixel\n```\n\nYou can also install modules you can manually downloaded, please check cli help for more information `piku add -h`.\n\nCurrently Piku just works for the Bundle 7, which was the most recent bundle when the tool was built.  But hopefully a full semver module index and supporting older versions and CircuitPython is something that can be done in the future.\n\n\n### A Complete Example with Adafruit Feather Sense\n\n##### Creating the Example Project\nAssuming you have successfully installed Piku, here is a complete example on how to create and deploy a Piku project to the [Adafruit Feather Sense](https://www.adafruit.com/product/4516) board using a Linux computer.\n\nFirst setup the board to user CircuitPython following the instructions found here:  \nhttps://learn.adafruit.com/adafruit-feather-sense/circuitpython-on-feather-sense\n\nAfter your board is setup and mounts as a `CIRCUITPY` USB drive create a new Piku project and enter the project directory.\n```\npiku create example\ncd example\n```\n\n##### Adding Main Example Program\nAfter you have created a project you will need to edit `project/code.py` which is the main file for your project (this is a CircuitPython convention). Open `project/code.py` and paste the following [Demo Code from AdaFruit](https://learn.adafruit.com/adafruit-feather-sense/circuitpython-sense-demo).\n\n```py\n# SPDX-FileCopyrightText: 2020 Kattni Rembor for Adafruit Industries\n#\n# SPDX-License-Identifier: MIT\n#\n"""Sensor demo for Adafruit Feather Sense. Prints data from each of the sensors."""\nimport time\nimport array\nimport math\nimport board\nimport audiobusio\nimport adafruit_apds9960.apds9960\nimport adafruit_bmp280\nimport adafruit_lis3mdl\nimport adafruit_lsm6ds.lsm6ds33\nimport adafruit_sht31d\n\ni2c = board.I2C()\n\napds9960 = adafruit_apds9960.apds9960.APDS9960(i2c)\nbmp280 = adafruit_bmp280.Adafruit_BMP280_I2C(i2c)\nlis3mdl = adafruit_lis3mdl.LIS3MDL(i2c)\nlsm6ds33 = adafruit_lsm6ds.lsm6ds33.LSM6DS33(i2c)\nsht31d = adafruit_sht31d.SHT31D(i2c)\nmicrophone = audiobusio.PDMIn(board.MICROPHONE_CLOCK, board.MICROPHONE_DATA,\n                              sample_rate=16000, bit_depth=16)\n\ndef normalized_rms(values):\n    minbuf = int(sum(values) / len(values))\n    return int(math.sqrt(sum(float(sample - minbuf) *\n                             (sample - minbuf) for sample in values) / len(values)))\n\napds9960.enable_proximity = True\napds9960.enable_color = True\n\n# Set this to sea level pressure in hectoPascals at your location for accurate altitude reading.\nbmp280.sea_level_pressure = 1013.25\n\nwhile True:\n    samples = array.array(\'H\', [0] * 160)\n    microphone.record(samples, len(samples))\n\n    print("\\nFeather Sense Sensor Demo")\n    print("---------------------------------------------")\n    print("Proximity:", apds9960.proximity)\n    print("Red: {}, Green: {}, Blue: {}, Clear: {}".format(*apds9960.color_data))\n    print("Temperature: {:.1f} C".format(bmp280.temperature))\n    print("Barometric pressure:", bmp280.pressure)\n    print("Altitude: {:.1f} m".format(bmp280.altitude))\n    print("Magnetic: {:.3f} {:.3f} {:.3f} uTesla".format(*lis3mdl.magnetic))\n    print("Acceleration: {:.2f} {:.2f} {:.2f} m/s^2".format(*lsm6ds33.acceleration))\n    print("Gyro: {:.2f} {:.2f} {:.2f} dps".format(*lsm6ds33.gyro))\n    print("Humidity: {:.1f} %".format(sht31d.relative_humidity))\n    print("Sound level:", normalized_rms(samples))\n    time.sleep(0.3)\n```\n\n##### Installing Dependencies\nNext install the required libraries for the AdaFruit Feather Sense example:\n```\npiku add adafruit_apds9960\npiku add adafruit_bmp280\npiku add adafruit_bus_device\npiku add adafruit_lis3mdl\npiku add adafruit_lsm6ds\npiku add adafruit_register\npiku add adafruit_sht31d\npiku add neopixel\n```\n\nThese modules should now found to your project `lib` folder, and your `pyproject.toml` file.  Confirm this by listing the files in your `lib` directory using `ls project/lib`. The ls command should return something the following if all modules were installed properly:\n```\nadafruit_apds9960  adafruit_bmp280.mpy  adafruit_bus_device  adafruit_lis3mdl.mpy  adafruit_lsm6ds  adafruit_register  adafruit_sht31d.mpy  neopixel.mpy\n```\n\nYour pyproject.toml file should now look something like this:\n```\n[tool.piku]\nproject = "example"\npiku = "0.1.1"\ncircuitpython = "7"\n\n[tool.piku.dependencies]\nadafruit_apds9960 = "~7"\nadafruit_bmp280 = "~7"\nadafruit_bus_device = "~7"\nadafruit_lis3mdl = "~7"\nadafruit_lsm6ds = "~7"\nadafruit_register = "~7"\nadafruit_sht31d = "~7"\nneopixel = "~7"\n```\n\n##### Deploying to the Device\nNow make sure your device is mounted as a USB drive and find the device\'s mount point.  This should be something like `/media/username/CIRCUITPY/` or a drive letter on windows. ***Make note of this!***\n\nAfter your device is connected and mounted, you can deploy your code using the deploy command:\n```\npiku deploy -d <your device path here>\n```\n\nNext before deploying confirm that the device selected is the correct device. When Piku deploys it first attempts to validate its a CircuitPython device and then tries to backup the contents of the device in a temporary location. After validation and backup it loads your files onto the device and removes almost all other files\n\nYou can also forego the `-d <your device path here>` argument and let Piku attempt to find your device, but please confirm that you are deploying to the correct device so you don\'t lost any data.\n\n##### Connecting to the Serial Port\nAfter your code is deployed you can connect to the serial port to see your code in action. It make take a minute for the device to reload. To connect to the serial port run:\n\n```\npiku serial\n```\n\nPiku will attempt to connect to the first available serial port and reach your device, if you have more than one serial port you may need to specify which port via the command like arguments.  Hit `ctrl-x` to exit or `ctrl-c`/`ctrl-d` to enter/exit the CircuitPython REPL.\n\nThanks it! Happy hacking!\n\n\n# Contributing\n\nContributions are very welcome, my time to work on the project is limited.  Please post issues and pull requests on Github if you would like to help forward the project.\n\n### Future Work\n\nIt would be great to support older versions of CircuitPython and properly detect and download modules using semver.\n\nMacOS support and testing. I don\'t own a Mac so there are some missing elements for MacOS support.\n\nLoading board specific examples when using create.  It would be neat to index or support examples for boards, so when you create a project the dependencies and `code.py` file are already filled in with starter code.\n',
    'author': 'Mark Raleson',
    'author_email': 'markraleson@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mraleson/piku.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
import shutil
import hashlib
import difflib
import requests


def nget(dictionary, keys, default=None):
    for key in keys[:-1]:
        dictionary = dictionary.get(key, {})
    return dictionary.get(keys[-1], default)

def nset(dictionary, keys, value):
    for key in keys[:-1]:
        dictionary = dictionary.setdefault(key, {})
    dictionary[keys[-1]] = value

def ndel(dictionary, keys):
    for key in keys[:-1]:
        dictionary = dictionary.setdefault(key, {})
    if keys[-1] in dictionary:
        del dictionary[keys[-1]]

def checksum(path):
    hasher = hashlib.md5()
    with open(path,'rb') as f:
        while chunk := f.read(128 * hasher.block_size):
            hasher.update(chunk)
    return hasher.digest()

def tree(path):
    paths = []
    for root, dirs, files in os.walk(path):
        paths.extend([os.path.relpath(os.path.join(root, d), path) for d in dirs])
        paths.extend([os.path.relpath(os.path.join(root, f), path) for f in files])
    return set(paths)

def copy(src, dst, recursive=True, contents=False):
    if os.path.isdir(src):
        if recursive:
            if contents:
                return shutil.copytree(src, dst, dirs_exist_ok=True)
            return shutil.copytree(src, os.path.join(dst, os.path.basename(src)), dirs_exist_ok=True)
        return os.makedirs(dst, exist_ok=True)
    return shutil.copy2(src, dst)

def remove(path, recursive=True):
    if os.path.isdir(path):
        if recursive:
            shutil.rmtree(path)
        else:
            try:
                os.rmdir(path)
            except OSError:
                pass # raised if dir is not empty, this should only happen when a file was ignored in the folder
    else:
        os.remove(path)

def similar(value, options):
    candidates = {k for k in options if value in k}
    return candidates.union(difflib.get_close_matches(value, options, 3, .3))

def download(url, path):
    with requests.get(url, stream=True) as r:
        with open(path, 'wb') as f:
            shutil.copyfileobj(r.raw, f)
    return path



from piku.core import config, modules


def add_command(args):
    # decode module
    module, type, path, version = modules.decode(args.module)

    # check if module already exists
    if config.get(f'tool.piku.dependencies.{module}'):
        print(f'Module {module} already added to project. To reinstall, remove the module and add it again.')
        return

    # suggest module if we can't find it in index
    if type == 'index' and not path:
        print(f'Unable to find source for requested module: {module}')
        suggestions = modules.suggest(module)
        if suggestions:
            print('Did you mean')
        for suggestion in suggestions:
            print(f' * {suggestion}')
        return

    # copy or download module
    print(f'Acquiring {module}...')
    if not modules.aquire(path):
        print(f'Unable to aquire requested module: {module}')
        return

    # save module in pyproject.toml
    config.set(f'tool.piku.dependencies.{module}', version)



import os
from piku.core import config, modules, utils


def remove_command(args):
    module = args.module.lower()

    # remove module from pyproject.toml
    if not config.remove(f'tool.piku.dependencies.{module}'):
        print(f'Unable to find matching module {module} in pyproject.toml')
        suggestions = utils.similar(module, config.get('tool.piku.dependencies').keys())
        if suggestions:
            print('Did you mean')
            for suggestion in suggestions:
                print(f' * {suggestion}')
        return

    # remove file from library
    if not modules.remove(module):
        print(f'Warning: Unable to find and delete matching file in source for module {module} in project')



try:
    from importlib import metadata
except ImportError: # for Python<3.8
    import importlib_metadata as metadata


def get_version():
    return metadata.version('piku')

def version_command(args):
    print(get_version())



from .add import add_command
from .deploy import deploy_command
from .info import info_command
from .install import install_command
from .create import create_command
from .remove import remove_command
from .serial import serial_command
from .version import version_command
